package fr.inria.indes.ggmm;

import fr.inria.bps.base.Fun1;
import fr.inria.bps.base.Liste;
import fr.inria.bps.base.Pair;
import fr.inria.indes.ggmm.expr.Ref;
import fr.inria.indes.ggmm.expr.Tid;
import fr.inria.indes.ggmm.mop.Rd;
import fr.inria.indes.ggmm.mop.Wr;
import fr.inria.lognet.sos.BooleanEvent;
import fr.inria.lognet.sos.ClientEvent;
import fr.inria.lognet.sos.Contour;
import fr.inria.lognet.sos.Event;
import fr.inria.lognet.sos.Shape;
import fr.inria.lognet.sos.Sos;
import fr.inria.lognet.sos.SosColor;
import fr.inria.lognet.sos.Viewer;
import fr.inria.lognet.sos.Wrapper;

public class ConfigView extends Wrapper {
    Liste<Tid> W = null;

    public ConfigView(final Configuration conf, Viewer view) {
	shape = Sos.border(5,
		Sos.column(2, new Shape[]{
			threadView(conf, view),
			Sos.row(0, Sos.row(10, storeView(conf), ownerView(conf), pipeView(conf, view)), eol()),
			eoc(90),
		}));

    }

    @SuppressWarnings("serial")
    Shape storeView(Configuration conf) {
	return(build(1, conf.S, new Fun1<Pair<Ref,Value>,Shape>() {
	    public Shape call(Pair<Ref,Value> slot) {
		return(Sos.label(slot.car + "=" + slot.cdr));
	    }}));
    }

    @SuppressWarnings("serial")
    Shape ownerView(Configuration conf) {
	return(build(1, conf.O, new Fun1<Pair<Ref,Tid>,Shape>() {
	    public Shape call(Pair<Ref,Tid> slot) {
		return(Sos.label(slot.car + "=" + slot.cdr));
	    }}));
    }

    /*
     * Temporary Store
     */
    @SuppressWarnings("serial")
    Shape pipeView(final Configuration conf, final Viewer view) {
	return(build(1, conf.sigma, new Fun1<Pair<Tid,Mop>,Shape>() {
	    public Shape call(final Pair<Tid,Mop> slot) {
		return(shapePipe(conf, view, slot, new ClientEvent() {
		    public void receive(int key, Event e) {
			view.push(conf.stepStoreR3(slot, true));
		    }		    		    
		},new ClientEvent() {
		    public void receive(int key, Event e) {
			view.push(conf.stepStore(slot, true));
		    }		    
		}));
	    }}));
    }

    Shape shapePipe(final Configuration conf, final Viewer view, final Pair<Tid,Mop> slot, ClientEvent c3, ClientEvent c) {
	Shape s = Sos.label(slot.car + ":" + slot.cdr, nextPipe(conf, slot));
	Shape ss = new Contour(1, Sos.border(2, Sos.left(s)));
	Shape st = Sos.label("");
	Mop mop = slot.cdr;
	if(mop instanceof Wr && ((Wr) mop).w != null) {
	    st = Sos.button("^", 0, new ClientEvent() {
		public void receive(int key, Event e) {
		    Visibility w = (W != null ? Visibility.finite(W) : Visibility.FULL);
		    //System.out.println("Try growing visibility to " + W);
		    view.push(conf.stepWriteEarly(slot, w, true));
		}
	    });
	} else if(slot.cdr instanceof Rd) {
	    st = Sos.button(Sos.label("R2", nextR3(conf, slot)), 0, c3);
	}
	return(Sos.row(0, st, Sos.button(Sos.reversedownexit(ss), 0, c), eol()));
    }

    SosColor nextPipe(Configuration conf, Pair<Tid,Mop> slot) {
	try {
	    Configuration r = conf.stepStore(slot, false);
	    if(r == conf) return(SosColor.black);
	    return(SosColor.blue);
	} catch(Exception e) {
	    return(SosColor.red);
	}
    }

    SosColor nextR3(Configuration conf, Pair<Tid,Mop> slot) {
	try {
	    Configuration r = conf.stepStoreR3(slot, false);
	    if(r == conf) return(SosColor.black);
	    return(SosColor.blue);
	} catch(Exception e) {
	    return(SosColor.red);
	}
    }

    /*
     * Threads
     */
    @SuppressWarnings("serial")
    Shape threadView(final Configuration conf, final Viewer view) {
	return(build(1, conf.T, new Fun1<Pair<Tid,Expression>,Shape>() {
	    public Shape call(final Pair<Tid,Expression> slot) {
		return(shapeThread(conf, slot, new ClientEvent() {
		    public void receive(int key, Event e) {
			view.push(conf.stepThread(slot, true));
		    }		    
		}, new ClientEvent() {
		    public void receive(int key, Event e) {
			view.push(conf.extrastepThread(slot, true));
		    }		    
		}));
	    }}));
    }

    Shape shapeThread(Configuration conf, Pair<Tid,Expression> slot, ClientEvent c, ClientEvent ec) {
	Shape s = Sos.label(slot.car + ":" + slot.cdr, nextThread(conf, slot));
	Shape ss = new Contour(1, Sos.border(2, Sos.left(s)));
	final Tid t = slot.car;
	Shape sel = Sos.checkbox(0, new ClientEvent() {
	    public void receive(int key, Event e) {
		if(((BooleanEvent) e).value) {
		    W = Liste.cons(t, W);
		} else {
		    W = Liste.delete(t, W);
		}
	    }
	});
	Shape extra = Sos.button(Sos.label("#f", extranextThread(conf, slot)), 0, ec);
	return(Sos.row(0, sel, extra, Sos.button(Sos.reversedownexit(ss), 0, c), eol()));
    }

    SosColor nextThread(Configuration conf, Pair<Tid,Expression> slot) {
	try {
	    Configuration r = conf.stepThread(slot, false);
	    if(r == conf) return(SosColor.black);
	    return(SosColor.blue);
	} catch(Exception e) {
	    return(SosColor.red);
	}
    }

    SosColor extranextThread(Configuration conf, Pair<Tid,Expression> slot) {
	try {
	    Configuration r = conf.extrastepThread(slot, false);
	    if(r == conf) return(SosColor.black);
	    return(SosColor.blue);
	} catch(Exception e) {
	    return(SosColor.red);
	}
    }

    public static <T> Shape build(int sp, Liste<T> l, Fun1<T,Shape> build1) {
	int n = Liste.length(l);
	Shape[] col = new Shape[n];
	for(int i=0 ; l!=null; l=l.cdr, i++)  col[i] = build1.call(l.car);
	return(Sos.column(0, Sos.column(sp, col), eoc(80)));
    }

    static Shape eol() {
	return(new Shape() {
	    public int wscale(boolean grow) {return(90);}
	    //public int hscale(boolean grow) {return(90);}
	});
    }

    static Shape eoc(final int r) {
	return(new Shape() {
	    //public int wscale(boolean grow) {return(90);}
	    public int hscale(boolean grow) {return(r);}
	});
    }
}
