package fr.inria.indes.ggmm;

import java.io.PrintStream;
import java.util.Hashtable;
import java.util.Vector;

import fr.inria.bps.base.Liste;
import fr.inria.bps.base.Pair;
import fr.inria.indes.ggmm.expr.Ref;
import fr.inria.indes.ggmm.expr.Tid;
import fr.inria.indes.ggmm.mop.Rd;
import fr.inria.indes.ggmm.mop.Wr;

public class Configuration {
    public static int counter = 0;
    static PrintStream trace = null;
    static {
	try {
	    //trace = new PrintStream("/tmp/trace");
	} catch(Exception err) {
	    System.out.println("Err " + err);
	}
    }
    
    // store
    public Liste<Pair<Ref,Value>> S;
    // ownership
    public Liste<Pair<Ref,Tid>> O;
    // Thread System
    public Liste<Pair<Tid,Expression>> T;
    // temporary store
    public Liste<Pair<Tid,Mop>> sigma;
    
    public Configuration(Liste<Pair<Ref,Value>> S, Liste<Pair<Ref,Tid>> O, Liste<Pair<Tid,Mop>> sigma, Liste<Pair<Tid,Expression>> T) {
	this .S = S;
	this.O = O;
	this.T = T;
	this.sigma = sigma;
    }

    public Configuration copy() {
	return(new Configuration(S, O, sigma, T));
    }
    
    public void setSC() {
	MemoryModel.current = MemoryModel.SC;
    }
    
    public void setNR() {
	MemoryModel.current = MemoryModel.NR;
    }
    
    public void setMR() {
	MemoryModel.current = MemoryModel.MR;
    }
 
    public Result direct() {
	Vector<Path> r = new Vector<Path>();
	Hashtable<String,Configuration> nodes = new Hashtable<String,Configuration>();
	long start = System.currentTimeMillis();
	dfrun(Liste.cons(this, null), nodes, r, true);
	long stop = System.currentTimeMillis();
	return(new Result("Direct Exhaustive", nodes.keySet().size(), r.toArray(new Path[]{}), (stop-start)));
    }
    
    public Result sliceAndDirect() {
	Liste<Configuration> l = itl(T, null, null, null);
	Vector<Path> allr = new Vector<Path>();
	Hashtable<String,Configuration> rescache = new Hashtable<String,Configuration>();
	int nbnodes = 0;
	int n = Liste.length(l);
	System.out.print(n + " ");
	long start = System.currentTimeMillis();
	for( ; l != null; l = l.cdr, n--) {
	    if((n % 1000) == 0) System.out.print(".");
	    Hashtable<String,Configuration> nodes = new Hashtable<String,Configuration>();
	    Vector<Path> r = new Vector<Path>();
	    l.car.dfrun(Liste.cons(l.car, null), nodes, r, true);
	    for( Path p : r ) {
		Configuration cres = p.path.car;
	    	if(rescache.get(cres.toString()) == null) {
	    	    rescache.put(cres.toString(), cres);
	    	    allr.add(p);
	    	}
	    }
	    nbnodes += nodes.keySet().size();
	}
	long stop = System.currentTimeMillis();
	System.out.println("Done");
	return(new Result("Direct on interleave", nbnodes, allr.toArray(new Path[]{}), (stop-start)));
    }

    public Liste<Configuration> refstep() {
	Liste<Configuration> r = null;
	Liste<Pair<Tid,Expression>> done = null;
	for(Liste<Pair<Tid,Expression>> l = T; l!=null; l=l.cdr) {
	    Configuration next = l.car.cdr.refstep(this, l.car.car, done, l.cdr);
	    done = new Liste<Pair<Tid,Expression>>(l.car, done);
	    if(next != this) r = new Liste<Configuration>(next, r);
	}
	return(r);
    }
    
    static int maxp = 0;
    static int done = 0;
    static int tera = 0;
    public void full_run() {
	maxp = 0;
	done = 0;
	tera = 0;
	full_dfrun(0);
	System.out.println("depth="+maxp+" tera="+tera);
    }

    public void full_dfrun(int path) {
	if(path > maxp) maxp = path;
	path++;
	for(Liste<Configuration> sons = step(false); sons != null; sons = sons.cdr) {
		done++;
		if(done == 1000000000) {
		    tera++;
		    done = 0;
		    System.out.println("depth="+maxp+" tera="+tera);
		}
		sons.car.full_dfrun(path);
	}
    }
     
    public Vector<Path> run(boolean fast) {
	Vector<Path> r = new Vector<Path>();
	Hashtable<String,Configuration> nodes = new Hashtable<String,Configuration>();
	dfrun(Liste.cons(this, null), nodes, r, fast);
	System.out.println(nodes.keySet().size() + " nodes, " + r.size() + " results.");
	return(r);
    }
    
    public Path[] interleaved() {
	Liste<Configuration> l = itl(T, null, null, null);
	Vector<Path> allr = new Vector<Path>();
	Hashtable<String,Configuration> rescache = new Hashtable<String,Configuration>();
	int n = Liste.length(l);
	for( ; l != null; l = l.cdr, n--) {
	    if((n % 1000) == 0) System.out.println(" " + n + " " + allr.size());
	    Hashtable<String,Configuration> nodes = new Hashtable<String,Configuration>();
	    Vector<Path> r = new Vector<Path>();
	    l.car.dfrun(Liste.cons(l.car, null), nodes, r, true);
	    for( Path p : r ) {
		Configuration cres = p.path.car;
	    	if(rescache.get(cres.toString()) == null) {
	    	    rescache.put(cres.toString(), cres);
	    	    allr.add(p);
	    	}
	    }
	}
	return(allr.toArray(new Path[]{}));
    }
  
    Liste<Configuration> itl(Liste<Pair<Tid,Expression>> l, Liste<Pair<Tid,Expression>> lcont ,Liste<Pair<Tid,Expression>> left, Liste<Configuration> done) {
	if(l == null) return((lcont == null) ? Liste.cons(this, done) : done);
	Configuration next = l.car.cdr.step(this, l.car.car, left, l.cdr);
	// CARE check that l.car.cdr is in normal form and thus not in stuck state.
	if(next == this) return(itl(l.cdr, lcont, Liste.cons(l.car, left), done));
	// CARE loose some speculative computation
	// check for an epsilon-reduction
	if(next.sigma == this.sigma) {
	    Liste<Pair<Tid,Expression>> ll = next.T;
	    // skip what it's done
	    for(Liste<Pair<Tid,Expression>> tt  = left; tt != null; tt = tt.cdr) ll = ll.cdr;
	    return(itl(ll, lcont, left, done));
	}
	return(itl(l.cdr, Liste.cons(l.car, lcont), Liste.cons(l.car, left), next.itl(next.T, null, null, done)));
    }

    public void dfrun(Liste<Configuration> path, Hashtable<String,Configuration> done, Vector<Path> res, boolean fast) {
	Liste<Configuration> sons = step(fast);
	if(sons == null) {
	    res.add(new Path(path));
	} else {
	    int n = done.keySet().size();
	    //if(n % 10000 == 0) System.out.println(" " + n);
	    if(n > 100000) dump(done);
	    for( ; sons != null; sons = sons.cdr) {
		Configuration conf = sons.car;
		String sconf = conf.toString();
		if(done.get(sconf) == null) {
		    if(trace != null)
			trace.println(""+conf.sigma);
		    done.put(sconf, conf);
		    conf.dfrun(Liste.cons(conf, path), done, res, fast);
		}
	    }
	}
    }
    
    void dump(Hashtable<String,Configuration> done) {
	try {
	    PrintStream out = new PrintStream("/tmp/dump");
	    for( String name : done.keySet() )
		out.println(done.get(name).toString());
	    out.close();
	} catch(Exception err) {
	    throw(new RuntimeException(err));
	}
	throw(new RuntimeException("Too large"));
    }
    
    public Path nextChoices() {
	return(new Path(step(false)));
    }
    
    public Configuration flushThreads() {
	// evaluate all the threads from left to right
	Configuration r = this;
	for(Configuration next = this; next!=null; next=next.nextThread()) r = next;
	return(r);
    }
    
    Configuration nextThread() {
	Liste<Pair<Tid,Expression>> done = null;
	for(Liste<Pair<Tid,Expression>> l = T; l!=null; l=l.cdr) {
	    Configuration next = l.car.cdr.step(this, l.car.car, done, l.cdr);
	    if(next != this) return(next);
	    // CARE loose some speculative computation
	    done = Liste.cons(l.car, done);
	}
	return(null);
    }
      
    public Liste<Configuration> step(boolean fast) {	
	Liste<Configuration> r = null;
	// Try all steps in expresssions
	Liste<Pair<Tid,Expression>> done = null;
	for(Liste<Pair<Tid,Expression>> l = T; l!=null; l=l.cdr) {
	    Configuration next = l.car.cdr.step(this, l.car.car, done, l.cdr);
	    if(next != this) r = Liste.cons(next, r);
	    next = l.car.cdr.extrastep(this, l.car.car, done, l.cdr);
	    if(next != this) r = Liste.cons(next, r);
	    done = Liste.cons(l.car, done);
	}
	if(fast && r != null) return(r);
	// Try all steps in Mops
	Liste<Pair<Tid,Mop>> mdone = null;
	for(Liste<Pair<Tid,Mop>> l = sigma; l!=null; l=l.cdr) {
	    Configuration next = l.car.cdr.step(this, l.car.car, mdone, l.cdr);
	    if(next != this) r = Liste.cons(next, r);
	    // try the R2 rule
	    if(l.car.cdr instanceof Rd) {
		Configuration nextR3 = ((Rd) l.car.cdr).step_read_early(this, l.car.car, mdone, l.cdr);
		if(nextR3 != this) r = Liste.cons(nextR3, r);
	    }
	    mdone = Liste.cons(l.car, mdone);
	}
	// Try to extend all visibility
	mdone = null;
	for(Liste<Pair<Tid,Mop>> l = sigma; l!=null; l=l.cdr) {
	    Mop mop = l.car.cdr;
	    if(mop instanceof Wr) {
		Wr wr = (Wr) mop;
		if(wr.w != null && !wr.w.isTop()) {
		    for(Liste<Liste<Tid>> wl = sof(null, Liste.domain(T), null); wl != null; wl=wl.cdr) {
			if(wl.car != null) {
			    if(MemoryModel.current.is_in_W(wl.car)) {
				Configuration next = wr.early(this, l.car.car, mdone, l.cdr, Visibility.finite(wl.car));
				if(next != this) r = Liste.cons(next, r);
			    }
			}
		    }
		    Configuration saturated = wr.early(this, l.car.car, mdone, l.cdr, Visibility.FULL);
		    if(saturated != this) r = Liste.cons(saturated, r);
		}
	    }
	    mdone = Liste.cons(l.car, mdone);
	}
	return(r);
    }
    
    static <T> Liste<Liste<T>> sof(Liste<T> p, Liste<T> l, Liste<Liste<T>> done) {
	if(l == null) return(Liste.cons(p, done));
	return(sof(Liste.cons(l.car, p), l.cdr, sof(p, l.cdr, done)));
    }
    
    public Configuration stepThread(Pair<Tid,Expression> cpu, boolean trace) {
	Liste<Pair<Tid,Expression>> done = null;
	Liste<Pair<Tid,Expression>> l = T;
	for(; l!=null && l.car != cpu; l=l.cdr) done = Liste.cons(l.car, done);
	boolean old = MemoryModel.debug;
	if(!trace) MemoryModel.debug = false;
	if(l != null) return(cpu.cdr.step(this, cpu.car, done, l.cdr));
	MemoryModel.debug = old;
	return(this);
    }
    
    public Configuration extrastepThread(Pair<Tid,Expression> cpu, boolean trace) {
	Liste<Pair<Tid,Expression>> done = null;
	Liste<Pair<Tid,Expression>> l = T;
	for(; l!=null && l.car != cpu; l=l.cdr) done = Liste.cons(l.car, done);
	boolean old = MemoryModel.debug;
	if(!trace) MemoryModel.debug = false;
	if(l != null) return(cpu.cdr.extrastep(this, cpu.car, done, l.cdr));
	MemoryModel.debug = old;
	return(this);
    }
    
    public Configuration stepStore(Pair<Tid,Mop> ins, boolean trace) {
	Liste<Pair<Tid,Mop>> done = null;
	Liste<Pair<Tid,Mop>> ts = sigma;
	for( ; ts !=null && ts.car != ins; ts = ts.cdr) done = Liste.cons(ts.car, done);
	boolean old = MemoryModel.debug;
	if(!trace) MemoryModel.debug = false;
	if(ts != null) return(ins.cdr.step(this, ins.car, done, ts.cdr));
	MemoryModel.debug = old;
	return(this);
    }
    
    public Configuration stepStoreR3(Pair<Tid,Mop> ins, boolean trace) {
	if(!(ins.cdr instanceof Rd)) return(this);
	Liste<Pair<Tid,Mop>> done = null;
	Liste<Pair<Tid,Mop>> ts = sigma;
	for( ; ts !=null && ts.car != ins; ts = ts.cdr) done = Liste.cons(ts.car, done);
	boolean old = MemoryModel.debug;
	if(!trace) MemoryModel.debug = false;
	if(ts != null) return(((Rd) ins.cdr).step_read_early(this, ins.car, done, ts.cdr));
	MemoryModel.debug = old;
	return(this);
    }
     
    public Configuration stepWriteEarly(Pair<Tid,Mop> ins, Visibility w, boolean trace) {
	if(!(ins.cdr instanceof Wr)) return(this);
	Liste<Pair<Tid,Mop>> done = null;
	Liste<Pair<Tid,Mop>> ts = sigma;
	for( ; ts !=null && ts.car != ins; ts = ts.cdr) done = Liste.cons(ts.car, done);
	boolean old = MemoryModel.debug;
	if(!trace) MemoryModel.debug = false;
	if(ts != null) return(((Wr) ins.cdr).early(this, ins.car, done, ts.cdr, w));
	MemoryModel.debug = old;
	return(this);
   }
    
    /*
    public boolean equals(Configuration conf) {
	if(conf == this) return(true);
	return(eqS(S, conf.S) && eqO(O, conf.O) && eqT(T, conf.T) && eqTS(sigma, conf.sigma));
    }
    
    public static boolean eqS(Liste<Pair<Ref,Val>> s1, Liste<Pair<Ref,Val>> s2) {
	if(s1 == s2) return(true);
	if(s1 == null || s2 == null) return(false);
	for( ; s1 == null || s2 == null; s1=s1.cdr, s2=s2.cdr)
	    if(s1.car.car != s2.car.car || !s1.car.cdr.equals(s2.car.cdr))
		return(false);
	return(s1 == s2);
    }
    
    public static boolean eqO(Liste<Pair<Ref,Tid>> o1, Liste<Pair<Ref,Tid>> o2) {
	if(o1 == o2) return(true);
	if(o1 == null || o2 == null) return(false);
	for( ; o1 == null || o2 == null; o1=o1.cdr, o2=o2.cdr)
	    if(o1.car.car != o2.car.car || o1.car.cdr != o2.car.cdr)
		return(false);
	return(o1 == o2);
    }
    
    public static boolean eqT(Liste<Pair<Tid,Expression>> te1, Liste<Pair<Tid,Expression>> te2) {
	if(te1 == te2) return(true);
	if(te1 == null || te2 == null) return(false);
	for( ; te1 == null || te2 == null; te1=te1.cdr, te2=te2.cdr)
	    if(te1.car.car != te2.car.car || !te1.car.cdr.equals(te2.car.cdr))
		return(false);
	return(te1 == te2);
    }
    
    public static boolean eqTS(Liste<Pair<Tid,Mop>> tm1, Liste<Pair<Tid,Mop>> tm2) {
	if(tm1 == tm2) return(true);
	if(tm1 == null || tm2 == null) return(false);
	for( ; tm1 == null || tm2 == null; tm1=tm1.cdr, tm2=tm2.cdr)
	    if(tm1.car.car != tm2.car.car || !tm1.car.cdr.equals(tm2.car.cdr))
		return(false);
	return(tm1 == tm2);
    }
   */
    
    public boolean equals(Object o) {
	//return((o instanceof Configuration) ? equals((Configuration) o) : false);
	// Quick and dirty
	return(toString().equals(o.toString()));
    }

    public int hashCode() {
	// return(S.hashCode() + O.hashCode() + T.hashCode() + sigma.hashCode());
	// Quick and dirty
	return(toString().hashCode());
    }
    
    public void dump(int level) {
	System.out.println(toString());	
    }

    public String toString() {
	return(T + "+" + sigma + "+" + S + "+" + O);
    }
}
