package fr.inria.indes.ggmm;

import java.util.Hashtable;
import java.util.Vector;

import fr.inria.bps.base.Liste;
import fr.inria.bps.base.Pair;
import fr.inria.indes.ggmm.expr.Affect;
import fr.inria.indes.ggmm.expr.Bar;
import fr.inria.indes.ggmm.expr.BarIsync;
import fr.inria.indes.ggmm.expr.BarLwsync;
import fr.inria.indes.ggmm.expr.BarSync;
import fr.inria.indes.ggmm.expr.BarWW;
import fr.inria.indes.ggmm.expr.Boolean;
import fr.inria.indes.ggmm.expr.DeRef;
import fr.inria.indes.ggmm.expr.If;
import fr.inria.indes.ggmm.expr.Ref;
import fr.inria.indes.ggmm.expr.Tid;
import fr.inria.indes.ggmm.expr.Var;
import fr.inria.indes.ggmm.expr.Void;

public class Examples {
    private static int g = 0;
   
    static String dmp(String[][] store) {
	String r = "";
	for(int i=0; i<store.length; i++) r = r + store[i][0] + "=" + store[i][1] + " ";
	return(r);
    }
     
    static boolean storeInclude(Liste<Pair<Ref,Value>> store, String[][] slots) {
	for( String[] slot : slots )
	    if(!storeHas(store, slot[0], slot[1])) return(false);
	return(true);
    }
    
    static boolean storeHas(Liste<Pair<Ref,Value>> store, String name, String value) {
	for(Liste<Pair<Ref,Value>> s = store; s != null; s = s.cdr)
	    if(name.equals(s.car.car.id))
		return(value.equals(s.car.cdr.toString()));	
	throw(new RuntimeException("store " + store + " not have the reference " + name));
    }
 
    /*
     * 
     * Basic expression and configuration contructors
     * 
     */
    static Configuration doconf(Expression[] ee) {
	Liste<Pair<Tid,Expression>> threads = null;
	Vector<Ref> store = new Vector<Ref>();
	for(int i=ee.length-1; i>=0; i--) {
	    threads = Liste.cons(Pair.cons(new Tid(i), ee[i]), threads);
	    ee[i].getRefs(store);
	}
	return(new Configuration(doStore(store), null, null, threads));
    }

    static Configuration doconf(Expression e0) {
	return(doconf(new Expression[]{e0}));
    }

    static Configuration doconf(Expression e0, Expression e1) {
	return(doconf(new Expression[]{e0, e1}));
    }

    static Configuration doconf(Expression e0, Expression e1, Expression e2) {
	return(doconf(new Expression[]{e0, e1, e2}));
    }

    static Configuration doconf(Expression e0, Expression e1, Expression e2, Expression e3) {
	return(doconf(new Expression[]{e0, e1, e2, e3}));
    }

    static Configuration doconf(Expression e0, Expression e1, Expression e2, Expression e3, Expression e4) {
	return(doconf(new Expression[]{e0, e1, e2, e3, e4}));
    }

    static Liste<Pair<Ref,Value>> doStore(Vector<Ref> refs) {
	Liste<Pair<Ref,Value>> r = null;
	for( Ref ref : refs )
	    r = Liste.cons(Pair.cons(ref, (Value) Boolean.ff), r);
	return(r);
    }
    
    static Hashtable<String,Ref> STD;
    
    static {
	String[] names = {"p","q", "r", "r0", "r1", "r2", "r3"};
	STD = new Hashtable<String,Ref>();
	for(int i=0; i<names.length; i++) STD.put(names[i], new Ref(names[i]));
    }
    
    static boolean isReg(String name) {
	return(name.length() > 1 && name.startsWith("r"));
    }

    static Boolean tt = Boolean.tt;
    static Boolean ff = Boolean.ff;
    static Void nil = Void.nil;;
    
    static Expression _if(Value test, Expression alors, Expression sinon) {
	return(new If(test, alors, sinon));
    }
    
    static Expression _if(DeRef r, Expression alors, Expression sinon) {
	Var v = new Var("t"+(g++));
	return(Expression.let(v, r, new If(v, alors, sinon)));
    }

    static Expression set(Ref r, Value v, boolean reg) {
	return(new Affect(r, v, reg));
    }

    static Expression set(Ref r, DeRef e, boolean reg) {
	Var v = new Var("t"+(g++));
	return(Expression.let(v, e, new Affect(r, v, reg)));
    }
    
    static Expression set(DeRef r, Value val) {
	Var v = new Var("t"+(g++));
	return(Expression.let(v, r, new Affect(v, val, false)));
    }
   
    static Expression set(String name, Value v) {
	return(set(STD.get(name), v, isReg(name)));
    }
    
    static Expression set(String name, String name2) {
	return(set(STD.get(name), STD.get(name2), isReg(name)));
    }
    
    static Expression set(String name, DeRef e) {
	return(set(STD.get(name), e, isReg(name)));
    }
  
    static DeRef deref(String name, String r) {
	return(new DeRef(name, STD.get(r)));
    }

    static Expression begin(Expression e1, Expression e2) {
	return(Expression.begin(e1, e2));
    }

    static Expression begin(Expression e1, Expression e2, Expression e3) {
	return(Expression.begin(e1, e2, e3));
    }
    
    /*
     * 
     * All configuration examples
     * 
     */
     public static Configuration EX_INTRO(Bar b1, Bar b2) { /* 1 */
	/*  p := tt  || q := tt
	 *  b1       || b2
	 *  r0 := !q || r1 := !p
	 */
	return(doconf(
		begin(set("p", tt), b1, set("r0", deref("n0", "q"))),
		begin(set("q", tt), b2, set("r1", deref("n1", "p"))) ));
    }
     
     public static Configuration EX_2() {
 	/*  p := tt  || r0 := !q
 	 *  q := tt  || r1 := !p
 	 */
 	return(doconf(
 		begin(set("p", tt), set("q", tt)),
 		begin(set("r0", deref("n0", "q")), set("r1", deref("n1", "p"))) ));
     }

     public static Configuration EX_LB(Bar b1, Bar b2) { /* 4 18 */
 	/*  r0 := !p  || r1 := !q
 	 *  b1        || b2
 	 *  q := tt   || p := tt
 	 */
 	return(doconf(
 		begin(set("r0", deref("n0", "p")), b1, set("q", tt)),
 		begin(set("r1", deref("n1", "q")), b2, set("p", tt)) ));
     }

     public static Configuration EX_THIN_AIR() { /* 5 */
 	/*  p := !q  || q := !p
 	 */
 	return(doconf(
 		set("p", deref("n0", "q")),
 		set("q", deref("n1", "p")) ));
     }

     public static Configuration EX_6() { /* 6 */
 	/*  p := tt  || q := tt
 	 *  r0 := !p || r2 := !q
 	 *  r1 := !q || r3 := !p
 	 */
 	return(doconf(
 		begin(set("p", tt), set("r0", deref("n0", "p")), set("r1", deref("n1", "q"))),
 		begin(set("q", tt), set("r2", deref("n2", "q")), set("r3", deref("n3", "p"))) ));
    }

     public static Configuration EX_7_ORI() {
 	/*  p := q  
 	 *  r0 := !p  || r1 := !q
 	 *  !r0 := tt || r2 := !p
 	 */
 	return(doconf(
 		begin(set("p", "q"), set("r0", deref("n", "p")), set(deref("n0", "r0"), tt)),
 		begin(set("r1", deref("n1", "q")), set("r2", deref("n2", "p"))) ));
    }

     public static Configuration EX_7() { /* 7 */
 	/*  p  := q  || r0 := !q
 	 *  !p := tt || r1 := !p
 	 */
 	return(doconf(
 		begin(set("p", "q"), set(deref("n", "p"), tt)),
 		begin(set("r0", deref("n0", "q")), set("r1", deref("n1", "p"))) ));
    }

     public static Configuration EX_WRC(Bar b1, Bar b2) { /* 8 15 */
 	/*  p := tt   || r0 := !p  || r1 := !q
 	 *            || b1        || b2
 	 *            || q := tt   || r2 := !p
 	 */
 	return(doconf(
 		set("p", tt),
 		begin(set("r0", deref("n0", "p")), b1, set("q", tt)),
 		begin(set("r1", deref("n1", "q")), b2, set("r2", deref("n2", "p"))) ));
    }

     public static Configuration EX_8() { /* 8 12 13 */
 	/*  p := tt  || r0 := !p  || r1 := !q
 	 *           || q := tt   || r2 := !p
 	 */
 	return(doconf(
 		set("p", tt),
 		begin(set("r0", deref("n0", "p")), set("q", tt)),
 		begin(set("r1", deref("n1", "q")), set("r2", deref("n2", "p"))) ));
    }
   
     public static Configuration EX_RWC(Bar b1, Bar b2) { /* 9  19 */
 	/*  p := tt  || r0 := !p  || q := tt 
 	 *           || b1        || b2
 	 *           || r1 := !q  || r2 := !p
 	 */
 	return(doconf(
 		set("p", tt),
 		begin(set("r0", deref("n0", "p")), b1, set("r1", deref("n1", "q"))),
 		begin(set("q", tt), b2, set("r2", deref("n2", "p"))) ));
    }
     
  
     public static Configuration EX_CC(Bar b1, Bar b2, Bar b3) { /* 10 */
 	/*  p := tt   || r0 := !p  || q := tt   || r2 := !p
 	 *            || b1        || b2        || b3
 	 *            || r1 := !q  || p := ()   || r3 := !p
 	 */
 	return(doconf(
 		set("p", tt),
 		begin(set("r0", deref("n0", "p")), b1, set("r1", deref("n1", "q"))),
 		begin(set("q", tt), b2, set("p", nil)),
 		begin(set("r2", deref("n2", "p")), b3, set("r3", deref("n3", "p"))) ));
    }

     public static Configuration EX_IRIW(Bar b1, Bar b2) { /* 11 14 */
 	/*  p := tt   || q := tt   || r0 := !p  || r2 := !q
 	 *                         || b1        || b2
 	 *                         || r1 := !q  || r3 := !p
 	 */
 	return(doconf(
 		set("p", tt),
 		set("q", tt),
 		begin(set("r0", deref("n0", "p")), b1, set("r1", deref("n1", "q"))),
 		begin(set("r2", deref("n2", "q")), b2, set("r3", deref("n3", "p"))) ));
    }

     
     // 
     // SPECULATION
     //
     
     
     //
     // APPENDIX B
     //   
    public static Configuration EX_MP(Bar b1, Bar b2) { /* 16 */
	/*  p := tt  || r0 := !q
	 *  b1       || b2
	 *  q := tt  || r1 := !p
	 */
	return(doconf( 
		begin(set("p", tt), b1, set("q", tt)),
		begin(set("r0", deref("n0", "q")), b2, set("r1", deref("n1", "p"))) ));
    }

    public static Configuration EX_SB(Bar b1, Bar b2) { /* 17 */
	/*  p := tt  || q := tt
	 *  b1       || b2
	 *  r0 := !q || r1 := !p
	 */
	return(doconf(
		begin(set("p", tt), b1, set("r0", deref("n0", "q"))),
		begin(set("q", tt), b2, set("r1", deref("n1", "p"))) ));
    }

    public static Configuration EX_ISA2(Bar b1, Bar b2, Bar b3) { /* 20 */
	/*   p := tt   || r0 := !q  || r1 := !r
	 *   b1        || b2        || b3  
	 *   q := tt   || r := tt   || r2 := !p
	 */
	return(doconf(
		begin(set("p", tt), b1, set("q", tt)),
		begin(set("r0", deref("n0", "q")), b2, set("r", tt)),
		begin(set("r1", deref("n1", "r")), b3, set("r2", deref("n2", "p"))) ));
   }

    public static Configuration EX_R(Bar b1, Bar b2) { /* 21 */
	/*  p := tt  || q := ff
	 *  b1       || b2
	 *  q := tt  || r0 := !p
	 */
	return(doconf(
		begin(set("p", tt), b1, set("q", tt)),
		begin(set("q", ff), b2, set("r0", deref("n0", "p"))) ));
    }

    public static Configuration EX_S(Bar b1, Bar b2) { /* 22 */
	/*  p := ff  || r0 := !q
	 *  b1       || b2
	 *  q := tt  || p := tt
	 */
	return(doconf(
		begin(set("p", ff), b1, set("q", tt)),
		begin(set("r0", deref("n0", "q")), b2, set("p", tt)) ));
    }

    public static Configuration EX_2_2W(Bar b1, Bar b2) {
	/*  p := ff  || q := ff
	 *  b1       || b2
	 *  q := tt  || p := tt
	 */
	return(doconf(
		begin(set("p", ff), b1, set("q", tt)),
		begin(set("q", ff), b2, set("p", tt)) ));
    }

    public static Configuration EX_WWC(Bar b1, Bar b2) {
	/*  p := tt  || r0 := !p  || r1 := !q
	 *           || b1        || b2
	 *           || q := tt   || p := ff
	 */
	return(doconf(
		set("p", tt),
		begin(set("r0", deref("n0", "p")), b1, set("q", tt)),
		begin(set("r1", deref("n1", "q")), b2, set("p", ff)) ));
   }

    public static Configuration EX_WRW_2W(Bar b1, Bar b2) {
	/*  p := ()  || r0 := !p  || q := ff
	 *           || b1        || b2
	 *           || q := tt   || p := tt
	 */
	return(doconf(
		set("p", nil),
		begin(set("r0", deref("n0", "p")), b1, set("q", tt)),
		begin(set("q", ff), b2, set("p", tt)) ));
   }

    public static Configuration EX_WRW_WR(Bar b1, Bar b2) {
	/*  p := tt  || r0 := !p  || q := ff
	 *           || b1        || b2
	 *           || q := tt   || r1 := !p
	 */
	return(doconf(
		set("p", tt),
		begin(set("r0", deref("n0", "p")), b1, set("q", tt)),
		begin(set("q", ff), b2, set("r1", deref("n1", "p"))) ));
   }

    public static Configuration EX_WRR_2W(Bar b1, Bar b2) {
	/*  p := ()  || r0 := !p  || q := tt 
	 *           || b1        || b2
	 *           || r1 := !q  ||p := tt
	 */
	return(doconf(
		set("p", nil),
		begin(set("r0", deref("n0", "p")), b1, set("r1", deref("n1", "q"))),
		begin(set("q", tt), b2, set("p", tt)) ));
   }

    public static Configuration EX_WRR_2W_r(Bar b1, Bar b2) {
	/*  p := ()  || r0 := !p  || q := tt 
	 *           || b1        || b2
	 *           || r1 := !q  ||p := tt
	 */
	return(doconf(
		begin(set("q", tt), b2, set("p", tt)),
		set("p", nil),
		begin(set("r0", deref("n0", "p")), b1, set("r1", deref("n1", "q"))) ));
   }

    public static Configuration EX_blw_w_006(Bar b1, Bar b2, Bar b3) {
	/*   p := tt   || q := ff   || r0 := !r
	 *   b1        || b2        || b3
	 *   q := tt   || r := tt   || r1 := !p
	 */
	return(doconf(
		begin(set("p", tt), b1, set("q", tt)),
		begin(set("q", ff), b2, set("r", tt)),
		begin(set("r0", deref("n0", "r")), b3, set("r1", deref("n1", "p"))) ));
   }

    /*
     * 
     * 
     * 
     */

    
    
    

    
    
    
    
 
    
    
    
    public static Configuration EX_18() {
	/*  p := tt  || r0 := !p  || r1 := !q
	 *           || lwsync    || lwsync
	 *           || q := tt   || r2 := !p
	 */
	return(doconf(
		set("p", tt),
		begin(set("r0", deref("n0", "p")), new BarLwsync(), set("q", tt)),
		begin(set("r1", deref("n1", "q")), new BarLwsync(), set("r2", deref("n2", "p"))) ));
   }
  
    public static Configuration EX_19() {
	return(EX_RWC(new BarLwsync(), new BarSync()));
   }

    public static Configuration EX_20() {
	/*  p := tt  || r0 := !p  || q := tt 
	 *           || sync      || lwsync
	 *           || r1 := !q  || r2 := !p
	 */
	return(doconf(
		set("p", tt),
		begin(set("r0", deref("n0", "p")), new BarSync(), set("r1", deref("n1", "q"))),
		begin(set("q", tt), new BarLwsync(), set("r2", deref("n2", "p"))) ));
   }

    public static Configuration EX_20_1() {
	/*  p := tt  || r0 := !p  || q := tt 
	 *           || sync      || sync
	 *           || r1 := !q  || r2 := !p
	 */
	return(doconf(
		set("p", tt),
		begin(set("r0", deref("n0", "p")), new BarSync(), set("r1", deref("n1", "q"))),
		begin(set("q", tt), new BarSync(), set("r2", deref("n2", "p"))) ));
   }

    public static Configuration EX_21() {
	return(EX_ISA2(new BarLwsync(), new BarLwsync(), new BarLwsync()));
    }

    public static Configuration EX_22() {
	return(EX_R(new BarLwsync(), new BarSync()));
    }

    public static Configuration EX_22_rev() {
	/*  p := tt  || q := ff
	 *  lwsync   || sync
	 *  q := tt  || r0 := !p
	 */
	return(doconf(
		begin(set("q", ff), new BarSync(), set("r0", deref("n0", "p"))),
		begin(set("p", tt), new BarLwsync(), set("q", tt)) ));
    }

    public static Configuration EX_22_1() {
	/*  p := tt  || q := ff
	 *  sync     || sync
	 *  q := tt  || r0 := !p
	 */
	return(doconf(
		begin(set("p", tt), new BarSync(), set("q", tt)),
		begin(set("q", ff), new BarSync(), set("r0", deref("n0", "p"))) ));
    }

    public static Configuration EX_22_2() {
	/*  p := tt  || q := ff
	 *  sync     || lwsync
	 *  q := tt  || r0 := !p
	 */
	return(doconf(
		begin(set("p", tt), new BarSync(), set("q", tt)),
		begin(set("q", ff), new BarLwsync(), set("r0", deref("n0", "p"))) ));
    }
    
    ///
    /// Speculative
    ///
    public static Configuration EX_IF_0() {
	/*  (if !p then r := ! p else ()) || p := tt
	 */
	return(doconf(
		_if(deref("n0", "p"), set("r", deref("n1", "p")), nil),
		set("p", tt) ));
    }
    
    public static Configuration EX_IF_1() {
	/*  (if !p then p:=tt else ())
	 */
	return(doconf(
		_if(deref("n0", "p"), set("p", tt), nil) ));
    }
    
    public static Configuration EX_IF_2() {
	/*  let x=!p in (if x then () else ()); (if x then () else ())
	 */
	Var v = new Var("t"+(g++));
	return(doconf(
		Expression.let(v, deref("n", "p"), begin(new If(v, nil, nil), new If(v, nil, nil))) ));
    }
    
    public static Configuration EX_MP_sync_ctrl() {
	/*  p := tt  ||
	 *  q := tt  || (if !q then r:=!p else ())
	 */
	return(doconf(
		begin(set("p", tt), new BarWW(), set("q", tt)),
		_if(deref("n0", "q"), set("r", deref("n1", "p")), set("r", nil)) ));
    }
    
    public static Configuration EX_MP_sync_ctrl_isync() {
	/*  p := tt  ||
	 *  q := tt  || (if !q then isync;r:=!p else ())
	 */
	return(doconf(
		begin(set("p", tt), new BarWW(), set("q", tt)),
		_if(deref("n0", "q"), begin(new BarIsync(), set("r", deref("n1", "p"))), set("r", nil)) ));
    }
    
    public static Configuration EX_PPOCA() {
	/*  r := tt  || (if !p then q := tt else ())
	 *  p := tt  || r0 := !q
	 *           || r1 := !r
	 */
	return(doconf(
		begin(set("r", tt), set("p", tt)),
		begin(_if(deref("n0", "p"), set("q", tt), nil),
			set("r0", deref("n1", "q")),
			set("r1", deref("n2", "r")) )));
    }
    
    public static Configuration EX_IF_PQQP() {
	/*  (if !p then q:=tt else ()) || (if !q then p:=tt else ())
	 */
	return(doconf(
		_if(deref("n0", "p"), set("q", tt), nil),
		_if(deref("n1", "q"), set("p", tt), nil) ));
    }
    
}
