package fr.inria.indes.ggmm;

import java.util.Vector;

import fr.inria.bps.base.Liste;
import fr.inria.bps.base.Pair;
import fr.inria.indes.ggmm.expr.Abs;
import fr.inria.indes.ggmm.expr.Call;
import fr.inria.indes.ggmm.expr.Ident;
import fr.inria.indes.ggmm.expr.MakeThread;
import fr.inria.indes.ggmm.expr.Ref;
import fr.inria.indes.ggmm.expr.Tid;
import fr.inria.indes.ggmm.expr.Var;
import fr.inria.indes.ggmm.expr.WithDo;

public abstract class Expression {
    
    /*
     *  Abstract methods
     */
    // Do one step of the Reference Operatinal Semantics (Figure 1)
    public abstract void dorefstep(StepResult res);

    // Do one step of M-Relaxed Operational Semantics of threads (Figure 2)
    public abstract void dostep(StepResult res);
    
    // Used for the beta reduction (see Call)
    public abstract Expression subst(Var v, Value e);
    
    public abstract Expression substIdent(Ident id, Value v);
    
    public abstract boolean hasIdent();

    public abstract void getRefs(Vector<Ref> v);

    
    // Specific for speculation
    public void doextrastep(StepResult res) {
	// must be overridden by If and all expression driving the context (call/Protected/...?)
    }
    
    // Specific for WithDo
    public boolean alreadyAcquire(WithDo child, Ref p) {
	// must be overridden by all expression driving the context (call/Protected/...?)
	return(false);
    }
    
    /*
     * Static methods constructing expressions
     */
    public static Expression begin(Expression e1, Expression e2) {
	Var dummy = new Var("_");
	return(new Call(new Abs(dummy, e2), e1));
    }
    
    public static Expression begin(Expression e1, Expression e2, Expression e3) {
	return(Expression.begin(e1, Expression.begin(e2, e3)));
    }
    
    public static Expression let(Var v, Expression vv, Expression body) {
	return(new Call(new Abs(v, body), vv));
    }

    public static Expression let(Var v1, Expression vv1, Var v2, Expression vv2, Expression body) {
	return(new Call(new Abs(v1, let(v2, vv2, body)), vv1));
    }

    public static Expression let(Var v1, Expression vv1, Var v2, Expression vv2, Var v3, Expression vv3, Expression body) {
	return(new Call(new Abs(v1, let(v2, vv2, v3, vv3, body)), vv1));
    }

    public static Expression let(Var v1, Expression vv1, Var v2, Expression vv2, Var v3, Expression vv3, Var v4, Expression vv4, Expression body) {
	return(new Call(new Abs(v1, let(v2, vv2, v3, vv3, v4, vv4, body)), vv1));
    }
    
    public static Expression par(Expression e1, Expression e2) {
	return(begin(new MakeThread(e2), e1));
    }
      
    /*
     * Step one expression in a Thread pool
     */
     public Configuration refstep(Configuration conf, Tid t, Liste<Pair<Tid, Expression>> done, Liste<Pair<Tid, Expression>> todo) {
	StepResult r = new StepResult(conf.copy(), t, this, done, todo);
	dorefstep(r);
	if(r.res == null) return(conf);
	r.conf.T = Liste.appendrev(done, Liste.cons(Pair.cons(t, r.res), r.right));
	return(r.conf);
    }
    
    public Configuration step(Configuration conf, Tid t, Liste<Pair<Tid, Expression>> done, Liste<Pair<Tid, Expression>> todo) {
	StepResult r = new StepResult(conf.copy(), t, this, done, todo);
	dostep(r);
	if(r.res == null) return(conf);
	r.conf.T = Liste.appendrev(done, Liste.cons(Pair.cons(t, r.res), r.right));
	return(r.conf);
    }
    
    public Configuration extrastep(Configuration conf, Tid t, Liste<Pair<Tid, Expression>> done, Liste<Pair<Tid, Expression>> todo) {
	StepResult r = new StepResult(conf.copy(), t, this, done, todo);
	doextrastep(r);
	if(r.res == null) return(conf);
	r.conf.T = Liste.appendrev(done, Liste.cons(Pair.cons(t, r.res), r.right));
	return(r.conf);
    }
  }
