package fr.inria.indes.ggmm;

import java.lang.reflect.Method;
import fr.inria.bps.base.Fun1;
import fr.inria.bps.base.Proc1;
import fr.inria.indes.ggmm.expr.Bar;
import fr.inria.indes.ggmm.expr.BarIsync;
import fr.inria.indes.ggmm.expr.BarLwsync;
import fr.inria.indes.ggmm.expr.BarPpo;
import fr.inria.indes.ggmm.expr.BarRR;
import fr.inria.indes.ggmm.expr.BarSync;
import fr.inria.indes.ggmm.expr.BarWR;
import fr.inria.indes.ggmm.expr.BarWW;
import fr.inria.lognet.sos.Shape;
import fr.inria.lognet.sos.Sos;
import fr.inria.lognet.sos.Viewer;

public class Main {
    
    static void usage() {
	System.out.println("Main [options] <NameofExample> {no | rr | wr | ww | ppo | sync | lwsync | isync}*");
	System.out.println(" where options are :");
	
	System.out.println("\t-h[elp]");
	System.out.println("\t\tprints the quick help and quit");
	
	System.out.println("\t-l[ist]");
	System.out.println("\t\tprints the names of available examples and quit");
	
	System.out.println("\t-m[odel] {sequential | noreordering | relaxed}");
	System.out.println("\t\tset the desired memory model, relaxed by default");

	System.out.println("\t-r[un] {i[nteractive] | ex[haustive]} | s[lice] | f[lush]");

	System.out.println("\t-c[heck] {d[ump] | n[bresults] <int> | a[llow] <Store> | f[orbid] <Store> }");
	System.out.println("\t\t<Store>={name=val{,name=val}*}");
	System.exit(0);
    }
 
    @SuppressWarnings("serial")
    public static void main(String[] args) {
	if(args.length < 1) usage();
	Configuration conf = null;
	Fun1<Configuration,Result> run = null;
	Proc1<Result> check = null;
	for(int i=0; i<args.length; i++) {
	    String cmd = args[i];
	    if(cmd.startsWith("-h")) {
		usage();
	    } else if(cmd.startsWith("-l")) {
		list();
	    } else if(cmd.startsWith("-m")) {
		String model = args[++i];
		if(model.toLowerCase().startsWith("r")) {
		    MemoryModel.current = MemoryModel.MR;
		} else if(model.toLowerCase().startsWith("s")) {
		    MemoryModel.current = MemoryModel.SC;
		} else if(model.toLowerCase().startsWith("n")) {
		    MemoryModel.current = MemoryModel.NR;
		} else {
		    throw(new RuntimeException("cannot find a memory model for " + model));
		}
	    } else if(cmd.startsWith("-r")) {
		String style = args[++i];
		if(style.toLowerCase().startsWith("i")) {
		    run = new Fun1<Configuration,Result>() {
			public Result call(Configuration conf) {
			    Sos.frame(new Viewer(conf, true, true), 10, 10, "It's all writes");
			    return(null);
			}
		    };
		} else if(style.toLowerCase().startsWith("e")) {
		    run = new Fun1<Configuration,Result>() {
			public Result call(Configuration conf) {
			    return(conf.direct());
			}
		    };
		} else if(style.toLowerCase().startsWith("s")) {
		    run = new Fun1<Configuration,Result>() {
			public Result call(Configuration conf) {
			    return(conf.sliceAndDirect());
			}
		    };
		} else if(style.toLowerCase().startsWith("f")) {
		    run = new Fun1<Configuration,Result>() {
			public Result call(Configuration conf) {
			    return(conf.flushThreads().direct());
			}
		    };
		} else {
		    throw(new RuntimeException("cannot run " + style));
		}
	    } else if(cmd.startsWith("-c")) {
		String style = args[++i];
		if(style.toLowerCase().startsWith("a")) {
		    final String[][] store = readStore(args[++i]);
		    check = new Proc1<Result>() {
			public void call(Result res) {
			    if(res.has(store)) System.out.println("OK\t"+res);
			    else System.out.println("FAIL\t"+res);
			}
		    };
		} else	if(style.toLowerCase().startsWith("f")) {
		    final String[][] store = readStore(args[++i]);
		    check = new Proc1<Result>() {
			public void call(Result res) {
			    if(res.has(store)) System.out.println("FAIL\t"+res);
			    else System.out.println("OK\t"+res);
			}
		    };
		} else	if(style.toLowerCase().startsWith("d")) {
		    check = new Proc1<Result>() {
			public void call(Result res) {
			    res.dump(1);
			}
		    };
		} else	if(style.toLowerCase().startsWith("n")) {
		    final int nbr = Integer.parseInt(args[++i]);
		    check = new Proc1<Result>() {
			public void call(Result res) {
			    if(res.results.length != nbr) System.out.println("FAIL\t"+res);
			    else System.out.println("OK\t"+res);
			}
		    };
 		} else {
		    throw(new RuntimeException("cannot check " + style));		    
		}
	    } else {
		Method m = getMethod("EX_"+cmd);
		int n = m.getParameterTypes().length;
		Bar[] params = new Bar[n];
		if(i+n+1 != args.length) {
		    System.out.println("There are a wrong number of arguments for applying " + cmd + " which need " + n + " barriers");
		    System.exit(-1);
		} else {
		    for(int j=0; j<n; j++)
			params[j] = decodeBar(args[i+j+1]);
		    i = args.length;
		}
		try {
		    conf = (Configuration) m.invoke(m, ((Object[]) params));
		} catch (Exception e) {
		    System.out.println("Something wrong when applying " + cmd);
		    System.exit(-1);
		}
	    }
	}
	System.out.print("====");
	for(int i=0; i<args.length; i++) System.out.print(" " + args[i]);
	System.out.println(" ====");
	if(run != null && conf != null) {
	    Result r = run.call(conf);
	    if(check != null && r != null) {
		check.call(r);
	    }
	}
    }
    
    static String[][] readStore(String buf) {
	String[] aff = buf.replaceAll("nil","()").split(",");
	int n = aff.length;
	String[][] r = new String[n][];
	for(int i=0; i<n; i++) r[i] = aff[i].split("=");
	return(r);
    }
    
    static Bar decodeBar(String name) {
	if(name.toLowerCase().startsWith("no")) return(new Bar());
	if(name.toLowerCase().startsWith("rr")) return(new BarRR());
	if(name.toLowerCase().startsWith("wr")) return(new BarWR());
	if(name.toLowerCase().startsWith("ww")) return(new BarWW());
	if(name.toLowerCase().startsWith("ppo")) return(new BarPpo());
	if(name.toLowerCase().startsWith("sy")) return(new BarSync());
	if(name.toLowerCase().startsWith("lw")) return(new BarLwsync());
	if(name.toLowerCase().startsWith("isy")) return(new BarIsync());
	throw(new RuntimeException("Cannot create a barrier with name " + name));	    
    }
 
    static Method getMethod(String name) {
	Class<Examples> c = Examples.class;
	Method[] methods = c.getDeclaredMethods();
	int n = methods.length;
	for(int i=0; i<n; i++) {
	    Method m = methods[i];
	    if(m.getName().equals(name))
		return(m);
	}
	throw(new RuntimeException("cannot find method " + name));
    }
    
    static void list() {
	Class<Examples> c = Examples.class;
	Method[] methods = c.getDeclaredMethods();
	int n = methods.length;
	for(int i=0; i<n; i++) {
	    Method m = methods[i];
	    Class<?>[] types = m.getParameterTypes();
	    String name = m.getName();
	    if(name.startsWith("EX_"))
		System.out.println(name.substring(3) + " need " + types.length + " barriers.");
	}
	System.exit(0);
    }

    /*
     * Add shape constructor for types Path and Configuration
     */
    static {
	Viewer.addConstructor("fr.inria.indes.ggmm.Main.path");	
	Viewer.addConstructor("fr.inria.indes.ggmm.Main.configuration");	
    }

    public static Shape path(Path path, Viewer view) {
	return(new PathView(path, view));
    }

    public static Shape configuration(Configuration conf, Viewer view) {
	return(new ConfigView(conf, view));
    }

}
