package fr.inria.indes.ggmm;

import fr.inria.bps.base.Liste;
import fr.inria.bps.base.Pair;
import fr.inria.indes.ggmm.expr.BarIsync;
import fr.inria.indes.ggmm.expr.BarLwsync;
import fr.inria.indes.ggmm.expr.BarPpo;
import fr.inria.indes.ggmm.expr.BarRR;
import fr.inria.indes.ggmm.expr.BarSync;
import fr.inria.indes.ggmm.expr.BarWR;
import fr.inria.indes.ggmm.expr.BarWW;
import fr.inria.indes.ggmm.expr.Ident;
import fr.inria.indes.ggmm.expr.Tid;
import fr.inria.indes.ggmm.mop.BarMop;
import fr.inria.indes.ggmm.mop.Guess;
import fr.inria.indes.ggmm.mop.JoinMop;
import fr.inria.indes.ggmm.mop.Rd;
import fr.inria.indes.ggmm.mop.RdMark;
import fr.inria.indes.ggmm.mop.SyncPlus;
import fr.inria.indes.ggmm.mop.Wr;

public abstract class MemoryModel {
    public static boolean debug = false;
    public String name;
    public static MemoryModel SC;
    public static MemoryModel NR;
    public static MemoryModel MR;

    public static MemoryModel current;

    public MemoryModel(String name) {
	this.name = name;
    }

    static {
	SC = new MemoryModel("Sequential Consistency") {
	    public boolean abs_commutability(Liste<Pair<Tid,Mop>> l, Pair<Tid,Mop> ins) {
		return(false);
	    }
	    public boolean is_in_W(Liste<Tid> l) {
		return(l == null || l.cdr == null);
	    }
	};
	NR = new MemoryModel("No Reordering           ") {
	    public boolean abs_commutability(Liste<Pair<Tid,Mop>> l, Pair<Tid,Mop> ins) {
		Tid t = ins.car;
		for( ; l != null; l = l.cdr)
		    if(l.car.car == t)
			return(false);
		return(true);
	    }
	    public boolean is_in_W(Liste<Tid> l) {
		return(true);
	    }
	};
	MR = new MemoryModel("Most Relaxed            ") {
	    public boolean abs_commutability(Liste<Pair<Tid,Mop>> l, Pair<Tid,Mop> ins) {
		return(true);
	    }
	    public boolean is_in_W(Liste<Tid> l) {
		return(true);
	    }
	};
	current = MR;
    }

    public String toString() {
	return(name);
    }

    public abstract boolean abs_commutability(Liste<Pair<Tid,Mop>> l, Pair<Tid,Mop> ins);

    public abstract boolean is_in_W(Liste<Tid> l);

    public static boolean A_lhd(Pair<Tid,Mop> ins1, Pair<Tid,Mop> ins2) {
	Tid tl = ins1.car;
	Tid tr = ins2.car;
	// We consider the case where the comparison is valid only on same thread.
	// if so the answer is is delegated to the operator on the left.
	if(debug) System.out.println(ins1 + " < " + ins2 + " strict ");
	if(tl == tr) if(ins1.cdr.strict(ins2.cdr)) return(true);
	// here are the cases to be considered on two different threads.
	if(debug) System.out.println(ins1 + " < " + ins2 + " ww ");
	Mop ml = ins1.cdr;
	Mop mr = ins2.cdr;
	//// interchange of writes
	if((ml instanceof Wr) && (mr instanceof Wr)) {
	    Wr wl = (Wr) ml;
	    Wr wr = (Wr) mr;
	    if(((wl.rho == wr.rho) || (wl.rho instanceof Ident)) && ((tl == tr) || Visibility.contains(wl.w, tr))) {
		//System.out.println("WW wl="+wl+" wr="+wr+" tl="+tl+" tr="+tr);
		return(true);
	    }
	}
	// t,wr^W_p,v < t',rd_p,v when t' in W
	if(debug) System.out.println(ins1 + " < " + ins2 + " wr ");
	if ((ml instanceof Wr) && (mr instanceof Rd)) {
	    Wr wl = (Wr) ml;
	    Rd rr = (Rd) mr;
	    if((wl.rho == rr.rho|| (wl.rho instanceof Ident)) && ((tl == tr) || Visibility.contains(wl.w , tr)))
		return(true);
	}
	if(debug) System.out.println(ins1 + " < " + ins2 + " wr (read done)");
	// Write Vs Old read
	if ((ml instanceof Wr) && (mr instanceof RdMark))
	    if(((Wr) ml).hasReaden(((RdMark) mr)))
		return(true);
	/// Never cross a join
	if(debug) System.out.println(ins1 + " < " + ins2 + " join ");
	if(mr instanceof JoinMop) if(tl != ((JoinMop) mr).tid) return(true);
	// cannot out a guess before an associated read mark
	if((ml instanceof RdMark) && (mr instanceof Guess) && (((RdMark) ml).from) == ((Guess) mr).orig) return(true);
	/// else
	if(debug) System.out.println(ins1 + " < " + ins2 + " ok ");
	return(false);
    }

    public boolean commutability(Configuration conf, Liste<Pair<Tid,Mop>> l, Pair<Tid,Mop> ins) {
	// Axiom (E)
	if(debug) System.out.println(ins + " Comm : E");
	if(l == null) return(true);
	if(debug) System.out.println(ins + " Comm : A");
	// Axiom (A_\lhd)
	for(Liste<Pair<Tid,Mop>> ll=l ; ll != null ; ll = ll.cdr) 
	    if(A_lhd(ll.car, ins))
		return(false);
	if(debug) System.out.println(ins + " Comm : Bar");
	// Barrier's special cases
	if(!bar_commutability(conf, l, ins)) return(false);
	// rest is delayed to an abstract method.
	if(debug) System.out.println(ins + " Comm : abs");
	return(abs_commutability(l, ins));
    }

    public boolean bar_lhd(Configuration conf, Pair<Tid,Mop> ins1, Pair<Tid,Mop> ins2) {
	Tid tl = ins1.car;
	Tid tr = ins2.car;
	Mop ml = ins1.cdr;
	Mop mr = ins2.cdr;
	// Basic cases where both intructions are from the same thread
	if(tl == tr) {
	    // PPO
	    if(BarPpo.isA(mr)) {
		if((ml instanceof Wr) && (!Visibility.isT(((Wr) ml).w))) return(true);
		if(ml instanceof Rd) return(true);
	    }
	    if(BarPpo.isA(ml)) {
		if(mr instanceof Wr) return(true);
		if(mr instanceof Rd) return(true);
	    }
	    // WR
	    if(BarWR.isA(mr) && (ml instanceof Wr)) return(true);
	    if(BarWR.isA(ml) && (mr instanceof Rd)) return(true);
	    // WW
	    if(BarWW.isA(mr) && (ml instanceof Wr)) return(true);
	    if(BarWW.isA(ml) && (mr instanceof Wr)) return(true);
	    // RR
	    if(BarRR.isA(mr) && (ml instanceof Rd)) return(true);
	    if(BarRR.isA(mr) && (ml instanceof RdMark)) return(true);
	    if(BarRR.isA(ml) && (mr instanceof Rd)) return(true);
	    // SYNC
	    if(BarSync.isA(mr) && ((ml instanceof Rd) || (ml instanceof Wr))) return(true);
	    if(BarSync.isA(mr) && (ml instanceof RdMark)) return(true);
	    if(BarSync.isA(ml) && ((mr instanceof Rd) || (mr instanceof Wr))) return(true);
	    // LWSYNC
	    if(BarLwsync.isA(mr) && (ml instanceof Rd)) return(true);
	    if(BarLwsync.isA(mr) && (ml instanceof RdMark)) return(true);
	    if(BarLwsync.isA(ml) && (mr instanceof Wr)) return(true);
	    // ISYNC
	    if(BarIsync.isA(mr) && (ml instanceof Guess)) return(true);
	    if(BarIsync.isA(ml) && (mr instanceof Rd)) return(true);
	}
	// the special case
	// SYNC
	if(BarLwsync.isA(mr) && (ml instanceof Wr) && ((tl == tr) || Visibility.contains(((Wr) ml).w, tr))) return(true);
	if(BarSync.isA(mr) && (ml instanceof Wr) && ((tl == tr) || Visibility.contains(((Wr) ml).w, tr))) return(true);
	// otherwise
	return(false);
    }

    boolean lwsync_comm(Liste<Pair<Tid,Mop>> l, Tid t) {
	for( ; l != null ; l = l.cdr) {
	    Tid tt = l.car.car;
	    Mop mop = l.car.cdr;
	    if((tt == t) && (mop instanceof BarMop) && (((BarMop) mop).bar instanceof BarLwsync)) {
		for(Liste<Pair<Tid,Mop>> ll = l.cdr; ll != null; ll=ll.cdr)
		    if((tt == ll.car.car) && ((ll.car.cdr instanceof Rd) || (ll.car.cdr instanceof RdMark)))
			return(true);
	    }
	}
	return(false);
    }

    public boolean bar_commutability(Configuration conf, Liste<Pair<Tid,Mop>> l, Pair<Tid,Mop> ins) {
	// Pairwise comparison
	for(Liste<Pair<Tid,Mop>> ll=l ; ll != null ; ll = ll.cdr) 
	    if(bar_lhd(conf, ll.car, ins))
		return(false);
	// Special cases
	Mop mop = ins.cdr;
	if((mop instanceof Rd) && lwsync_comm(l, ins.car)) return(false);
	// else
	return(true);
    }

    public boolean restricted_commutability(Configuration conf, Liste<Pair<Tid,Mop>> l, Pair<Tid,Mop> ins) {
	return(commutability(conf, onlySyncPlus(l), ins));
    }

    static Liste<Pair<Tid,Mop>> onlySyncPlus(Liste<Pair<Tid,Mop>> l) {
	Liste<Pair<Tid,Mop>> r = null;
	for( ; l != null; l=l.cdr)
	    if(l.car.cdr instanceof SyncPlus)
		r = Liste.cons(l.car, r);
	return(r);
    }

}
