package fr.inria.indes.ggmm;

import fr.inria.bps.base.Liste;
import fr.inria.bps.base.Pair;
import fr.inria.indes.ggmm.expr.Abs;
import fr.inria.indes.ggmm.expr.Call;
import fr.inria.indes.ggmm.expr.If;
import fr.inria.indes.ggmm.expr.MakeRef;
import fr.inria.indes.ggmm.expr.Ref;
import fr.inria.indes.ggmm.expr.Tid;
import fr.inria.indes.ggmm.expr.Var;
import fr.inria.indes.ggmm.expr.Boolean;
import fr.inria.indes.ggmm.mop.Wr;

public class Recette {
    static MemoryModel model;
    
    public static void main(String[] args) {
	check1(beta1(), rbeta1());
	check1(if1(), rif1());
	check1(if2(), rif2());
	check1(mref(), rmref());
	System.out.println("Done.");
    }
    
    static void check1(Configuration b, Configuration e) {
	Liste<Configuration> l = b.step(false);
	if(l == null) {
	    System.out.println("** The configuration has no small step successor");
	    System.out.println(" " + b);
	    return;
	}
	if(l.cdr != null) {
	    System.out.println("** The configuration must have only one small step successor instead of " + Liste.length(l));
	    System.out.println(" " + b);
	    return;
	}
	if(!l.car.toString().equals(e.toString())) {
	    System.out.println("** Result mismatch");
	    System.out.println("computed:" + l.car);
	    System.out.println("needed:" + e);
	}
    }

    static Configuration doconf(Expression e) {
	Tid t = new Tid(0);
	return(new Configuration(null, null, null, Liste.cons(Pair.cons(t, e), null)));
    }

    static Configuration doconf(Expression e, Mop m) {
	Tid t = new Tid(0);
	return(new Configuration(null, null, Liste.cons(Pair.cons(t,m), null), Liste.cons(Pair.cons(t, e), null)));
    }
 
    //
    static Configuration beta1() {
	Var v = new Var("v");
	return(doconf(new Call(new Abs(v, v), Boolean.ff)));
    }
    
    static Configuration rbeta1() {
	return(doconf(Boolean.ff));
    }
    
    //
    static Configuration if1() {
	return(doconf(new If(Boolean.tt, Boolean.tt, Boolean.ff)));
    }
    
    static Configuration rif1() {
	return(doconf(Boolean.tt));
    }
    
    static Configuration if2() {
	return(doconf(new If(Boolean.ff, Boolean.tt, Boolean.ff)));
    }
    
    static Configuration rif2() {
	return(doconf(Boolean.ff));
    }
    
    //
    static Configuration mref() {
	return(doconf(new MakeRef(Boolean.tt)));
    }
    
    static Configuration rmref() {
	Ref res = new Ref(0);
	return(doconf(res, new Wr(res,Boolean.tt)));
    }
}
