package fr.inria.indes.ggmm;

import fr.inria.bps.base.Liste;
import fr.inria.indes.ggmm.expr.Tid;

public abstract class Visibility {
    public static final Visibility EMPTY = noVisibility();
    public static final Visibility FULL = fullVisibility();

    static Visibility noVisibility() {
	return(new Visibility() {
	    public boolean has(Tid t) {
		return(false);
	    }
	    public boolean lessSingleton(Tid t) {
		return(true);
	    }
	    public boolean isTop() {
		return(false);
	    }
	    public Liste<Tid> getList() {
		return(null);
	    }
	    public String toString() {
		return("");
	    }
	});
    }
    
    static Visibility fullVisibility() {
	return(new Visibility() {
	    public boolean has(Tid t) {
		return(true);
	    }
	    public boolean lessSingleton(Tid t) {
		return(false);
	    }
	    public boolean isTop() {
		return(true);
	    }
	    public Liste<Tid> getList() {
		throw(new RuntimeException("Cannot getList of Top"));
	    }
	    public String toString() {
		return("T");
	    }
	});	
    }

    public abstract boolean has(Tid t); // Liste.indexOf(t, this.list) > 0

    public abstract boolean lessSingleton(Tid t); // this include or equal {t}

    public abstract boolean isTop();

    public abstract Liste<Tid> getList();

    public static boolean isT(Visibility w) {
	if(w == null) return(false);
	return(w.isTop());	
    }
    
    public static boolean contains(Visibility w, Tid t) {
	if(w == null) return(false);
	return(w.has(t));
    }

    public static boolean include(Visibility w1, Visibility w2) {
	if(w2.isTop()) return(!w1.isTop());
	if(w1.isTop()) return(false);
	return(Liste.include(w1.getList(), w2.getList()));
    }

    public static Visibility finite(final Liste<Tid> l) {
	if(l == null) throw(new RuntimeException("at least a singleton"));
	if(l.cdr == null) {
	    return(new Visibility() {
		public boolean has(Tid t) {
		    return(l.car == t);
		}	
		public boolean lessSingleton(Tid t) {
		    return(l.car == t);
		}
		public boolean isTop() {
		    return(false);
		}
		public Liste<Tid> getList() {
		    return(l);
		}
		public String toString() {
		    return(""+l);
		}	    
	    });
	} else {
	    return(new Visibility() {
		public boolean has(Tid t) {
		    return(Liste.indexOf(t, l) >= 0);
		}	
		public boolean lessSingleton(Tid t) {
		    return(false);
		}
		public boolean isTop() {
		    return(false);
		}
		public Liste<Tid> getList() {
		    return(l);
		}
		public String toString() {
		    return(""+l);
		}	    
	    });
	}
    }   
}
