package fr.inria.indes.ggmm.expr;

import java.util.Vector;

import fr.inria.indes.ggmm.Expression;
import fr.inria.indes.ggmm.StepResult;
import fr.inria.indes.ggmm.Value;


public class Call extends Expression {
    public Value fun;
    public Expression arg;
    
    public Call(Value fun, Expression arg) {
	this.fun = fun;
	this.arg = arg;
    }
    
    public boolean hasIdent() {
	return(fun.hasIdent() || arg.hasIdent());
    }

    public void getRefs(Vector<Ref> v) {
	fun.getRefs(v);
	arg.getRefs(v);
    }

    public Expression subst(Var v, Value e) {
	return(new Call((Value) fun.subst(v, e), arg.subst(v, e)));
    }

    public Expression substIdent(Ident id, Value v) {
	return(new Call((Value) fun.substIdent(id, v), arg.substIdent(id, v)));
  }

    public void dorefstep(StepResult r) {
	if(!(arg instanceof Value)) {
	    arg.dorefstep(r);
	    if(r.res == null) return;
	    r.res = new Call(fun, r.res);
	    return;
	}
	if(!(fun instanceof Abs)) {
	    r.kind = StepResult.Kind.STUCK_NOTFUN;
	    return;
	}
	// Redex
	Abs abs = (Abs) fun;
	r.res = abs.body.subst(abs.var, ((Value) arg));
    }
    
    public void dostep(StepResult r) {
	if(!(arg instanceof Value)) {
	    arg.dostep(r);
	    if(r.res == null) return;
	    r.res = new Call(fun, r.res);
	    return;
	}
	if(!(fun instanceof Abs)) {
	    r.kind = StepResult.Kind.STUCK_NOTFUN;
	    return;
	}
	// Redex
	Abs abs = (Abs) fun;
	r.res = abs.body.subst(abs.var, ((Value) arg));
    }
    
    public void doextrastep(StepResult r) {
	if(!(arg instanceof Value)) {
	    arg.doextrastep(r);
	    if(r.res == null) return;
	    r.res = new Call(fun, r.res);
	    return;
	}
    }

    public boolean alreadyAcquire(WithDo child, Ref p) {
	return(arg.alreadyAcquire(child, p));
    }
 
    public String toString() {
	if(fun instanceof Abs) {
	    Abs f = (Abs) fun;
	    if(f.var.name.equals("_")) return(arg + "; " + f.body);
	    return("let " + f.var + " = " + arg + " in " + f.body);
	}
	return("(" + fun + " " + arg + ")");
    }
}
