package fr.inria.indes.ggmm.expr;

import java.util.Vector;

import fr.inria.indes.ggmm.Expression;
import fr.inria.indes.ggmm.StepResult;
import fr.inria.indes.ggmm.Value;
import fr.inria.indes.ggmm.mop.Guess;

public class If extends Expression {
    public Value test;
    public Expression alors;
    public Expression sinon;
    
    public If(Value test, Expression alors, Expression sinon) {
	this.test = test;
	this.alors = alors;
	this.sinon = sinon;
    }
 
    public boolean hasIdent() {
	return(test.hasIdent() || alors.hasIdent() || sinon.hasIdent());
    }

    public void getRefs(Vector<Ref> v) {
	test.getRefs(v);
	alors.getRefs(v);
	sinon.getRefs(v);
    }
  
    public Expression subst(Var v, Value e) {
	return(new If((Value) test.subst(v, e), alors.subst(v, e), sinon.subst(v, e)));
    }

    public Expression substIdent(Ident id, Value v) {
	return(new If((Value) test.substIdent(id, v), alors.substIdent(id, v), sinon.substIdent(id, v)));
    }

    public void dorefstep(StepResult r) {
	if(!(test instanceof Boolean)) {
	    r.kind = StepResult.Kind.STUCK_NOTBOOL;
	    return;
	}
	if(test == Boolean.ff) {
	    r.res = sinon;
	    return;
	} 
	r.res = alors;
    }    
    
    public static boolean speculative = true;
    public void dostep(StepResult r) {
	dorefstep(r);
	if(speculative && r.res == null && test instanceof Ident) {
	    // Speculative computation
	    r.addPipe(new Guess((Ident) test, Boolean.tt));
	    r.res = alors.substIdent((Ident) test, Boolean.tt);
	}
    }
    
    public void doextrastep(StepResult r) {
	if(speculative && r.res == null && test instanceof Ident) {
	    // Speculative computation
	    r.addPipe(new Guess((Ident) test, Boolean.ff));
	    r.res = sinon.substIdent((Ident) test, Boolean.ff);
	}	
    }
  
    public String toString() {
	return("(if " + test + " then " + alors + " else " + sinon + ")");
    }
}
