package fr.inria.indes.ggmm.expr;

import java.util.Vector;

import fr.inria.bps.base.Liste;
import fr.inria.bps.base.Pair;
import fr.inria.indes.ggmm.Expression;
import fr.inria.indes.ggmm.StepResult;
import fr.inria.indes.ggmm.Value;
import fr.inria.indes.ggmm.mop.JoinMop;

public class Join extends Expression {
    Value value;
    
    public Join(Value value) {
	this.value = value;
    }

    public boolean hasIdent() {
	return(value.hasIdent());
    }

    public void getRefs(Vector<Ref> v) {
	value.getRefs(v);
    }
  
    public Expression subst(Var v, Value e) {
	return(new Join((Value) value.subst(v, e)));
    }

    public Expression substIdent(Ident id, Value v) {
	return(new Join((Value) value.substIdent(id, v)));
    }

    public void dostep(StepResult r) {
	if(!(value instanceof Tid)) {
	    r.kind = StepResult.Kind.STUCK_NOTTHREAD;
	    return;
	}
	Tid t = (Tid) value;
	if(nwait(r, r.left, t)) return;
	if(nwait(r, r.right, t)) return;
	if(t == r.t) return;
	r.kind = StepResult.Kind.WAIT_JOIN;	
    }
 
    boolean nwait(StepResult r, Liste<Pair<Tid, Expression>> l, Tid t) {
	Expression e = Liste.cassq(l, t);
	if(e != null) {
	    if(e instanceof Value && !e.hasIdent()) {
		r.addPipe(new JoinMop(t));
		r.res = Void.nil;
	    } else
		r.kind = StepResult.Kind.WAIT_JOIN;
	    return(true);
	}
	return(false);
    }

    public void dorefstep(StepResult r) {
	if(!(value instanceof Tid)) {
	    r.kind = StepResult.Kind.STUCK_NOTTHREAD;
	    return;
	}
	Tid t = (Tid) value;
	if(wait(r, r.left, t)) return;
	if(wait(r, r.right, t)) return;
	if(t == r.t) return;
	throw(new RuntimeException("Invalid thread " + t));
    }
    
    boolean wait(StepResult r, Liste<Pair<Tid, Expression>> l, Tid t) {
	Expression e = Liste.cassq(l, t);
	if(e != null) {
	    if(e instanceof Value)
		r.res = Void.nil;
	    else
		r.kind = StepResult.Kind.WAIT_JOIN;
	    return(true);
	}
	return(false);
    }

    public String toString() {
	return("(join " + value + ")");
    }
}
