package fr.inria.indes.ggmm.expr;

import java.util.Vector;

import fr.inria.bps.base.Liste;
import fr.inria.bps.base.Pair;
import fr.inria.indes.ggmm.Expression;
import fr.inria.indes.ggmm.StepResult;
import fr.inria.indes.ggmm.Value;
import fr.inria.indes.ggmm.mop.Acquire;

public class WithDo extends Expression {
    public Value lock;
    public Expression todo;
    
    public WithDo(Value lock, Expression todo) {
	this.lock = lock;
	this.todo = todo;
    }
    
    public boolean hasIdent() {
	return(lock.hasIdent() || todo.hasIdent());
    }

    public void getRefs(Vector<Ref> v) {
	lock.getRefs(v);
	todo.getRefs(v);
    }
  
    public Expression subst(Var v, Value e) {
	return(new WithDo((Value) lock.subst(v, e), todo.subst(v, e)));
    }

    public Expression substIdent(Ident id, Value v) {
	return(new WithDo((Value) lock.substIdent(id, v), todo.substIdent(id, v)));
   }

    public void dostep(StepResult r) {
	if(!(lock instanceof Ref)) {
	    r.kind = StepResult.Kind.STUCK_NOTREF;
	    return;
	}
	Ref p = (Ref) lock;
	if(r.todo.alreadyAcquire(this, p)) {
	    r.res = todo;
	    return;
	}
	r.addPipe(new Acquire(p));
	r.res = new Protected(todo, p);
    }
    
    public void dorefstep(StepResult r) {
	if(!(lock instanceof Ref)) {
	    r.kind = StepResult.Kind.STUCK_NOTREF;
	    return;
	}
	Ref p = (Ref) lock;
	Tid own = Liste.cassq(r.conf.O, p);
	if(own == null) {
	    r.conf.O = new Liste<Pair<Ref, Tid>>(new Pair<Ref, Tid>(p, r.t), r.conf.O);
	    r.res = new Protected(todo, p);
	    return;
	}
	if(own == r.t) {
	    r.res = todo;
	    return;
	}
	r.kind = StepResult.Kind.WAIT_WITH;
	return;
    }

    public boolean alreadyAcquire(WithDo child, Ref p) {
	if(child == this) return(false);
	if(p == lock) return(true);
	return(todo.alreadyAcquire(child, p));
    }
   
    public String toString() {
	return("(with " + lock + " do " + todo + ")");
    }
}
