package fr.inria.indes.ggmm.mop;

import fr.inria.bps.base.Liste;
import fr.inria.bps.base.Pair;
import fr.inria.indes.ggmm.Configuration;
import fr.inria.indes.ggmm.MemoryModel;
import fr.inria.indes.ggmm.Mop;
import fr.inria.indes.ggmm.MopConfig;
import fr.inria.indes.ggmm.Value;
import fr.inria.indes.ggmm.expr.Ident;
import fr.inria.indes.ggmm.expr.Ref;
import fr.inria.indes.ggmm.expr.Tid;

public class Rd extends Mop {
    public Value rho;
    public Ident v;

    public Rd(Value rho, Ident v) {
	this.rho = rho;
	this.v = v;
    }

    public Mop substIdent(Ident id, Value val) {
	return(new Rd((rho == id) ? val : rho, v));
    }

    public void step(MopConfig r) {
	if(MemoryModel.current.commutability(r.conf, r.before, Pair.cons(r.thread, (Mop)this))) {
	    // Rule read (R1)
	    if(!(rho instanceof Ref)) throw(new RuntimeException("Internal error"));
	    Ref p = (Ref) rho;
	    Ident nu = (Ident) v;
	    Value val = Liste.cassq(r.conf.S, p);
	    if(val == null) throw(new RuntimeException("Invalid store " + rho));
	    r.substIdent(nu, val);
	}
    }

    public Configuration step_read_early(Configuration conf, Tid t, Liste<Pair<Tid, Mop>> after, Liste<Pair<Tid, Mop>> before) {
	MopConfig r = new MopConfig(conf.copy(), t, this, after, before);
	try_read_early(r);
	if(!r.done) return(conf);
	r.conf.sigma = r.res;
	return(r.conf);
    }

    public void try_read_early(MopConfig r) {
	if(v instanceof Ident) {
	    // Try to rule R2 (read early) 
	    Ident nu = (Ident) v;
	    Liste<Pair<Tid,Mop>> l = r.before;
	    Liste<Pair<Tid,Mop>> d1 = null;
	    for( ; l != null ; l=l.cdr) {
		if(is_read_early(r, l.cdr, l.car, d1)) {
		    r.substIdent(nu, ((Wr) l.car.cdr).v);
		    return;
		}
		d1 = Liste.cons(l.car, d1);
	    }
	}
    }

    boolean is_read_early(MopConfig r, Liste<Pair<Tid,Mop>>delta0, Pair<Tid,Mop> top, Liste<Pair<Tid,Mop>> delta1) {
	Mop mop = top.cdr;
	if(!(mop instanceof Wr)) return(false);
	Wr wr = (Wr) mop;
	if(wr.rho != rho) return(false);
	if(!MemoryModel.current.commutability(r.conf, delta1, Pair.cons(r.thread, (Mop) this))) return(false);
	if(!MemoryModel.current.restricted_commutability(r.conf, delta0, Pair.cons(r.thread, (Mop) this))) return(false);
	//if(!MemoryModel.current.commutability(r.conf, delta0, top)) return(false);
	if(!wr.w.has(r.thread)) return(false);
	RdMark mark = new RdMark(v);
	r.after = Liste.cons(Pair.cons(r.thread, (Mop) mark), r.after);
	r.before = Liste.appendrev(delta1, Liste.cons(Pair.cons(top.car, (Mop) wr.addReaden(mark)), delta0));
	return(true);
    }

    public boolean strict(Mop mop) {
	if(mop instanceof Wr) {
	    Wr wr = (Wr) mop;
	    return((rho == wr.rho) || (rho instanceof Ident));
	}
	if(mop instanceof Release) {
	    Release r = (Release) mop;
	    return((rho == r.ref) || (rho instanceof Ident));

	}
	if(mop instanceof Spw) return(true);
	return(false);	
    }

    public String toString() {
	return("read_" + rho + "," + v);
    }
}
