package fr.inria.indes.ggmm.mop;

import fr.inria.bps.base.Liste;
import fr.inria.bps.base.Pair;
import fr.inria.indes.ggmm.Configuration;
import fr.inria.indes.ggmm.Expression;
import fr.inria.indes.ggmm.MemoryModel;
import fr.inria.indes.ggmm.Mop;
import fr.inria.indes.ggmm.MopConfig;
import fr.inria.indes.ggmm.Value;
import fr.inria.indes.ggmm.Visibility;
import fr.inria.indes.ggmm.expr.Ident;
import fr.inria.indes.ggmm.expr.Ref;
import fr.inria.indes.ggmm.expr.Tid;

public class Wr extends Mop {
    public Value rho;
    public Value v;
    public Visibility w;
    public Liste<RdMark> readen;
    
    public Wr(Value rho, Value v, Visibility w, Liste<RdMark> readen) {
	this.rho = rho;
	this.v = v;
	this.w = w;
	this.readen = readen;
   }
    
    public Wr(Value rho, Value v) {
	this(rho, v, Visibility.EMPTY, null);
    }

    public Wr addReaden(RdMark m) {
	return(new Wr(rho, v, w, Liste.cons(m, readen)));
    }
    
    public boolean hasReaden(RdMark m) {
	return(Liste.indexOf(m, readen) >= 0);
    }
    
    public Mop substIdent(Ident id, Value v) {
	return(new Wr((rho == id) ? v : rho, (this.v == id) ? v : this.v, w, readen));
    }

    public void step(MopConfig r) {
	if((!(v.hasIdent())) && MemoryModel.current.commutability(r.conf, r.before, Pair.cons(r.thread, (Mop)this))) {
	    // Try rule R4 (write)
	    Liste<Pair<Tid,Mop>> l = r.before;
	    Liste<Pair<Tid,Mop>> d1 = null;
	    for( ; l != null ; l=l.cdr) {
		if(sameWrite(r, l.cdr, l.car, d1)) return;
		d1 = Liste.cons(l.car, d1);
	    }
	    if(!(rho instanceof Ref)) return;
	    // Gotcha
	    r.conf.S = Liste.setassq(r.conf.S, (Ref) rho, v);
	    r.commit();
	    return;
	}
    }
    
    public Configuration early(Configuration conf, Tid t, Liste<Pair<Tid, Mop>> after, Liste<Pair<Tid, Mop>> before, Visibility wp) {
	//System.out.println("early " + t + ": " + w + " -> " + wp);
//	if((MemoryModel.current.commutability(conf, before, Pair.cons(t, (Mop)this))) && (wp.has(t)) && Visibility.include(w, wp) && !Visibility.include(wp, w) 
//		 & allusefull(t, wp, conf.T, after)) {
		if(true && (wp.has(t)) && Visibility.include(w, wp) && !Visibility.include(wp, w) 
			&& allusefull(t, wp, conf.T, after)) {
//	if(true && (wp.has(t)) && Visibility.include(w, wp) && !Visibility.include(wp, w) // Idea of Gerard on 24.02
//			&& allusefull(t, wp, conf.T, after)) {
	    //System.out.println("write early ok");
	    Configuration nconf = conf.copy();
	    nconf.sigma = Liste.appendrev(after, Liste.cons(Pair.cons(t,(Mop)new Wr(rho, v, wp, readen)), before));
	    return(nconf);
	}
	//System.out.println("write early failed");
	return(conf);
    }
    
    public static boolean early_optim = true;
    static boolean allusefull(Tid me, Visibility wp, Liste<Pair<Tid,Expression>> threads, Liste<Pair<Tid,Mop>> after) {
	if(!early_optim) return(true);
	if(wp.isTop()) return(true);
	for(Liste<Tid> l = wp.getList(); l != null; l = l.cdr)
	    if(l.car != me)
		if(!live(l.car, threads))
		    if(!hasRead(l.car, after))
			return(false);
	return(true);
    }
    
    static boolean live(Tid t, Liste<Pair<Tid,Expression>> threads) {
	Expression e = Liste.cassq(threads, t);
	if(e == null) throw(new RuntimeException("internal error : someone delete a thread " + t));
	if(e instanceof Value) return(false);
	return(true);
    }
    
    static boolean hasRead(Tid t, Liste<Pair<Tid,Mop>> l) {
	for( ; l != null; l = l.cdr)
	    if(l.car.car == t && (l.car.cdr instanceof Rd))
		return(true);
	return(false);
    }
 
    boolean sameWrite(MopConfig conf, Liste<Pair<Tid,Mop>>delta0, Pair<Tid,Mop> top, Liste<Pair<Tid,Mop>> delta1) {
	Mop mop = top.cdr;
	if(!(mop instanceof Wr)) return(false);
	Wr wr = (Wr) mop;
	if((!(wr.rho instanceof Ident)) && wr.rho != rho) return(false);
	if(wr.w == null) return(false);
	if(wr.w.lessSingleton(top.car)) return(false);
	return(true);
    }

    public boolean strict(Mop mop) {
	if(mop instanceof Rd) {
	    Rd rd = (Rd) mop;
	    return((rho == rd.rho) || (rho instanceof Ident));
	}
	if(mop instanceof RdMark) {
	    RdMark rd = (RdMark) mop;
	    return(Liste.indexOf(rd, readen) >= 0);
	}
	if(mop instanceof Release) {
	    Release r = (Release) mop;
	    return((rho == r.ref) || (rho instanceof Ident));
	}
	if(mop instanceof Spw) return(true);
	return(false);
    }

    public String toString() {
	return("Wr" + (w==null ? "" : ("^"+w))+ "_" + rho + "," + v);
    }
}
